/*
 * $QNXLicenseC:
 * Copyright 2019, QNX Software Systems.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You
 * may not reproduce, modify or distribute this software except in
 * compliance with the License. You may obtain a copy of the License
 * at: http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as
 * contributors under the License or as licensors under other terms.
 * Please review this entire file for other proprietary rights or license
 * notices, as well as the QNX Development Suite License Guide at
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */

/*
 * Modfications copyright (c) 2019-2021, Texas Instruments Incorporated
 *
 */

#include "tisci_mgr.h"
#include "proto.h"

#define TISCI_DEVICE_NAME       "/dev/tisci"

#define SPROXY_RECV_TIMEOUT     1000
#define SCI_MAX_MESSAGE_SIZE    2048
static int          verbose = 0;

static int sci_io_devctl(resmgr_context_t *ctp, io_devctl_t *msg, RESMGR_OCB_T *ocb);
static int ti_sci_msg_xfer(void *ibuf, void *obuf, int *nbytes);

uint64_t Sciclient_qnxVirtToPhyFxn(const void *virtAddr,
                                   void *appData);

void * Sciclient_qnxPhyToVirtFxn(uint64_t phyAddr,
                                 void *appData);

static void getRevision(void)
{
    int32_t status = CSL_PASS;

    /* Fill in version request message */
    struct tisci_msg_version_req request;
    const Sciclient_ReqPrm_t      reqPrm =
    {
        TISCI_MSG_VERSION,
        TISCI_MSG_FLAG_AOP,
        (uint8_t *) &request,
        sizeof(request),
        SCICLIENT_SERVICE_WAIT_FOREVER
    };

    struct tisci_msg_version_resp response;
    Sciclient_RespPrm_t           respPrm =
    {
        0,
        (uint8_t *) &response,
        sizeof (response)
    };

    /* Request version */
    status = Sciclient_service_rsmgr(&reqPrm, &respPrm);
    if (CSL_PASS == status)
    {
        if (respPrm.flags == TISCI_MSG_FLAG_ACK)
        {
            status = CSL_PASS;
            sci_sloginfo(" SYSFW Firmware Version %s", (char *) response.str);
            sci_sloginfo(" SYSFW Firmware revision 0x%x", response.version);
            sci_sloginfo(" SYSFW ABI revision %d.%d", response.abi_major, response.abi_minor);
        }
        else
        {
            sci_slogerr("%s (%d): DMSC Firmware Get Version failed\n", __FUNCTION__,__LINE__);
        }
    }
    else
    {
        sci_slogerr("%s (%d): DMSC Firmware Get Version failed\n", __FUNCTION__,__LINE__);
    }
}

int main(int argc, char *argv[])
{
    struct stat             sbuf;
    int         id;
    int         option;
    int32_t status;
    resmgr_connect_funcs_t connect_funcs;
    resmgr_io_funcs_t io_funcs;
    dispatch_t *dpp;
    resmgr_attr_t rattr;
    dispatch_context_t *ctp;
    iofunc_attr_t ioattr;
    Sciclient_ConfigPrms_t config;

    /* Only allow one instance */
    if (-1 != stat(TISCI_DEVICE_NAME, &sbuf))
    {
        perror("tisci resmgr already running...");
        return (-1);
    }

    /* Get IO priveleges */
    if (ThreadCtl(_NTO_TCTL_IO, NULL) == -1) {
        perror("ThreadCtl(_NTO_TCTL_IO");
        return 1;
    }

    /* Initialize SCI config */
    Sciclient_configPrmsInit(&config);

    /* Initialize SCI */
    status = Sciclient_init(&config);
    if (status != 0)
    {
        perror("SCI INitialization failed\n");
        sci_slogerr("%s: Failed to intiialize SCI Client", argv[0]);
    }

    getRevision();

    /* Initialize the dispatch interface */
    dpp = dispatch_create();
    if (!dpp) {
        sci_slogerr("%s: Failed to create dispatch interface", argv[0]);
        return (errno);
    }

    /* Initialize the resource manager attributes */
    memset(&rattr, 0, sizeof(rattr));

    /* Initialize the connect functions */
    iofunc_func_init(_RESMGR_CONNECT_NFUNCS, &connect_funcs, _RESMGR_IO_NFUNCS, &io_funcs);
    io_funcs.devctl = sci_io_devctl;
    iofunc_attr_init(&ioattr, S_IFCHR | 0644, NULL, NULL);

    /* Attach the device name */
    id = resmgr_attach(dpp, &rattr, TISCI_DEVICE_NAME, _FTYPE_ANY, 0, &connect_funcs, &io_funcs, &ioattr);
    if (id == -1) {
        sci_slogerr("%s: Failed to attach pathname", argv[0]);
        return (errno);
    }

    /* Allocate a context structure */
    ctp = dispatch_context_alloc(dpp);

    while ( (option = getopt(argc, argv, "v")) != -1) {
        switch (option) {

            case 'v':
                verbose++;
                break;
            default:
                fprintf(stderr,"Unsupported option '-%c'\n",option);
                goto fail0;
        }
    }


    /* Run in the background */
    if (procmgr_daemon(EXIT_SUCCESS, PROCMGR_DAEMON_NOCLOSE | PROCMGR_DAEMON_NODEVNULL ) == -1) {
        sci_slogerr("%s: procmgr_daemon", argv[0]);
        goto fail0;
    }

    while (1) {
        if ((ctp = dispatch_block(ctp)) == NULL) {
            sci_slogerr("%s: Block error", argv[0]);
            goto fail0;
        }
        dispatch_handler(ctp);
    }

fail0:
    return (-errno);
}

static int ti_sci_msg_xfer(void *ibuf, void *obuf, int *nbytes)
{
    int32_t status = CSL_PASS;
    tisci_msg_t *tisci_msg_in = (tisci_msg_t *) ibuf;
    tisci_msg_t *tisci_msg_out = (tisci_msg_t *) obuf;

    Sciclient_ReqPrm_t   *reqFromClient = (Sciclient_ReqPrm_t *) &tisci_msg_in->reqPrm;
    Sciclient_RespPrm_t   *respFromClient = (Sciclient_RespPrm_t *) &tisci_msg_in->respPrm;
    Sciclient_RespPrm_t  *respToClient = (Sciclient_RespPrm_t *) &tisci_msg_out->respPrm;

    /* Change the physical address of the client response to virtual mapping, so we can
     * write directly into the address space provided by user.  This somewhat removes the
     * need for the response portion of the message to the user.  Need to think through
     * if the full response has to go back to user, or just the payload.
     */

    uint64_t *pReqVirtAddr = 0;
    uint64_t *pRespVirtAddr = 0;
    if((reqFromClient->pReqPayload != 0) && (reqFromClient->reqPayloadSize > 0))
    {
        pReqVirtAddr = Sciclient_qnxPhyToVirtFxn((uint64_t) reqFromClient->pReqPayload, &reqFromClient->reqPayloadSize);
        if(pReqVirtAddr == 0)
        {
            sci_slogerr("%s (%d): sciclient service call unable to map memory\n",__FUNCTION__,__LINE__);
            status = CSL_EALLOC;
        }
        reqFromClient->pReqPayload = (uint8_t *) pReqVirtAddr;
    }
    if((respFromClient->pRespPayload != 0) && (respFromClient->respPayloadSize > 0))
    {
        pRespVirtAddr = Sciclient_qnxPhyToVirtFxn((uint64_t) respFromClient->pRespPayload, &respFromClient->respPayloadSize);
        if(pRespVirtAddr == 0)
        {
            sci_slogerr("%s (%d): sciclient service call unable to map memory\n",__FUNCTION__,__LINE__);
            status = CSL_EALLOC;
        }
        respFromClient->pRespPayload = (uint8_t *) pRespVirtAddr;
    }

    /* Send request to sciclient library */
    if (CSL_PASS == status)
    {
        status = Sciclient_service_rsmgr(reqFromClient, respFromClient);
    }

    if (CSL_PASS == status)
    {
        *nbytes = sizeof(tisci_msg_t);
        if ((respFromClient->flags & TISCI_MSG_FLAG_ACK) == TISCI_MSG_FLAG_ACK)
        {
            /* Change the virtual address of the request and response virtual address,
             * to physical */
            if((reqFromClient->pReqPayload != 0) && (reqFromClient->reqPayloadSize > 0))
            {
                reqFromClient->pReqPayload = (uint8_t *) Sciclient_qnxVirtToPhyFxn((uint64_t *)
                    reqFromClient->pReqPayload, &reqFromClient->reqPayloadSize);
            }
            if((respFromClient->pRespPayload != 0) && (respFromClient->respPayloadSize > 0))
            {
                respFromClient->pRespPayload = (uint8_t *) Sciclient_qnxVirtToPhyFxn((uint64_t *)
                    respFromClient->pRespPayload, &respFromClient->respPayloadSize);
            }
        }
        else
        {
            status = CSL_EFAIL;
            sci_slogerr("%s (%d): sciclient service call failed on ACK\n",__FUNCTION__,__LINE__);

            /* Change the virtual address of the request and response virtual address,
             * to physical */
            if((reqFromClient->pReqPayload != 0) && (reqFromClient->reqPayloadSize > 0))
            {
                if (verbose) {
                    sci_sloginfo("%s: translate request\n",__func__);
                }
                reqFromClient->pReqPayload =
                    (uint8_t *) Sciclient_qnxVirtToPhyFxn((uint64_t *) reqFromClient->pReqPayload,
                     &reqFromClient->reqPayloadSize);
            }
            if((respFromClient->pRespPayload != 0) && (respFromClient->respPayloadSize > 0))
            {
                if (verbose) {
                    sci_sloginfo("%s: translate response\n",__func__);
                }
                respFromClient->pRespPayload =
                    (uint8_t *) Sciclient_qnxVirtToPhyFxn((uint64_t *) respFromClient->pRespPayload,
                     &respFromClient->respPayloadSize);
            }
        }
    }
    else {
          sci_slogerr("%s (%d): Call to Sciclient_service_rsmgr failed\n",__func__,__LINE__);
    }

    if (verbose) {
        sci_sloginfo("%s: respToClient->flags = 0x%08x\n",__func__, respToClient->flags);
        sci_sloginfo("%s: respFromClient->flags = 0x%08x\n",__func__, respFromClient->flags);
    }
    respToClient->flags = respFromClient->flags;
    respToClient->respPayloadSize = respFromClient->respPayloadSize;
    respToClient->pRespPayload = respFromClient->pRespPayload;
    if (verbose) {
        sci_sloginfo("%s: respToClient->flags = 0x%08x\n",__func__, respToClient->flags);
        sci_sloginfo("%s: respFromClient->flags = 0x%08x\n",__func__, respFromClient->flags);
        sci_sloginfo("%s: respToClient->payloadSize = 0x%d\n",__func__, respToClient->respPayloadSize);

    }

    if(pReqVirtAddr != 0)
    {
        if(munmap_device_memory(pReqVirtAddr, reqFromClient->reqPayloadSize) == -1)
        {
          sci_slogerr("%s (%d): Memory unmap failed\n",__func__,__LINE__);
        }
    }
    if(pRespVirtAddr != 0)
    {
        if(munmap_device_memory(pRespVirtAddr, respFromClient->respPayloadSize) == -1)
        {
          sci_slogerr("%s (%d): Memory unmap failed\n",__func__,__LINE__);
        }
    }

    return (EOK);
}

static int sci_io_devctl(resmgr_context_t *ctp, io_devctl_t *msg, RESMGR_OCB_T *ocb)
{
    int     status, nbytes;
    int     err = EOK;

    if ((status = iofunc_devctl_default(ctp, msg, ocb)) != _RESMGR_DEFAULT) {
        return status;
    }

    nbytes = 0;
    if (msg->i.dcmd == DCMD_TISCI_MESSAGE) {
        if (verbose) {
            sci_sloginfo("%s: DCMD_TISCI_MESSAGE rxd\n",__func__);
        }
        /* Should also pass msg->o here, to copy in response message, re-use user's buffer */
        err = ti_sci_msg_xfer(_DEVCTL_DATA(msg->i), _DEVCTL_DATA(msg->o), &nbytes);
    } else {
        err = EINVAL;
    }

    if (nbytes == 0 || err != EOK) {
        printf("%s: ERROR nbytes/%d err/%d EOK/%d\n",__func__,nbytes,err,EOK);
        sci_slogerr("%s (%d): ERROR nbytes/%d err/%d EOK/%d\n",__FUNCTION__,__LINE__,nbytes,err,EOK);
        return (err);
    }

    msg->o.ret_val = 0;
    msg->o.nbytes = nbytes;
    if (verbose) {
        sci_sloginfo("%s: sizeof(msg->o)/%d sizeof(tisci_msg_t)/%d nbytes/%d\n",__func__, (int)sizeof(msg->o) , (int)sizeof(tisci_msg_t), nbytes);
    }
    return (_RESMGR_PTR(ctp, &msg->o, sizeof(msg->o) + sizeof(tisci_msg_t)));
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL$ $Rev$")
#endif
