/*
 * $QNXLicenseC:
 * Copyright 2019, QNX Software Systems.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"). You
 * may not reproduce, modify or distribute this software except in
 * compliance with the License. You may obtain a copy of the License
 * at: http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OF ANY KIND, either express or implied.
 *
 * This file may contain contributions from others, either as
 * contributors under the License or as licensors under other terms.
 * Please review this entire file for other proprietary rights or license
 * notices, as well as the QNX Development Suite License Guide at
 * http://licensing.qnx.com/license-guide/ for other information.
 * $
 */


/*
 * TI J721E-EVM
 */
#include "startup.h"
#include "ti_sci.h"
#include "drivers_usbwrapper.h"
#include "J721E_pinmux.h"
#include <hw/omap_i2c.h>
#include "board.h"

extern int r5_display;

#define J721E_CTRLMMR_PADCONFIG 0x11C000
#define J721E_WKUP_CTRLMMR_PADCONFIG    (0x43000000 + 0x1C000)

#define J721E_GPIO0_BASE        0x600000
#define J721E_GPIO_DIR(x)       (0x10 + ((x) / 32) * 0x28)
#define J721E_GPIO_SET_DATA(x)  (0x18 + ((x) / 32) * 0x28)
#define J721E_GPIO_BIT(x)       (1 << ((x) % 32))

#define USB0_WRAPPER_BASE   0x6000000
#define USB0_SUBSYS_BASE    0x4104000
#define USB1_WRAPPER_BASE   0x6400000
#define USB1_SUBSYS_BASE    0x4114000

#define sysFreqKhz  19200

static int init_usb_host(int host, int usb2mode)
{
    uint32_t    dev_id;
    paddr_t     usbss_adr, usb_wrapper_adr;

    if (host == 0) {
        dev_id = TISCI_DEV_USB0;
        usbss_adr = USB0_SUBSYS_BASE;
        usb_wrapper_adr = USB0_WRAPPER_BASE;
        // set pinmux, need datashhet for mux config
        out32(J721E_CTRLMMR_PADCONFIG + PIN_USB0_DRVVBUS, PIN_PULL_DISABLE | PIN_MODE(0));  // VBUS control
    } else {
        dev_id = TISCI_DEV_USB1;
        usbss_adr = USB1_SUBSYS_BASE;
        usb_wrapper_adr = USB1_WRAPPER_BASE;
        // set pinmux, need datashhet for mux config
        out32(J721E_CTRLMMR_PADCONFIG + PIN_MCAN1_TX, PIN_PULL_DISABLE | PIN_MODE(4));  // VBUS control
    }
    if (ti_sci_set_device_state(dev_id, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: HOST %d, failed! %d\n", __func__, host, __LINE__);
        return (-1);
    }

    // TODO!!!
    // should move to board specific xhci driver
    Drivers_SetUsbPhyConfig(usbss_adr, sysFreqKhz);
    Drivers_SetUsbMode(usbss_adr, usb_wrapper_adr, USB_MODE_HOST, usb2mode);

    // USB sub-system clock, parent clock
    if (ti_sci_cmd_set_clk_parent(dev_id, TISCI_DEV_USB0_USB2_REFCLOCK_CLK,
            TISCI_DEV_USB0_USB2_REFCLOCK_CLK_PARENT_GLUELOGIC_HFOSC0_CLKOUT) == -1) {
        kprintf("%s: HOST %d, failed! %d\n", __func__, host, __LINE__);
        return (-1);
    }

    if (ti_sci_set_clock_state(dev_id, TISCI_DEV_USB0_USB2_REFCLOCK_CLK,
            0, MSG_CLOCK_SW_STATE_AUTO) == -1) {
        kprintf("%s: HOST %d, failed! %d\n", __func__, host, __LINE__);
        return (-1);
    }

    if (ti_sci_set_clock_state(dev_id, TISCI_DEV_USB0_CLK_LPM_CLK,
            0, MSG_CLOCK_SW_STATE_AUTO) == -1) {
        kprintf("%s: HOST %d, failed! %d\n", __func__, host, __LINE__);
        return (-1);
    }

    return (0);
}

static void init_i2c(void)
{
    // pinmux for I2C0
    out32(J721E_CTRLMMR_PADCONFIG + PIN_I2C0_SCL, PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_I2C0_SDA, PIN_INPUT_ENABLE | PIN_MODE(0));

    // pinmux for I2C1
    out32(J721E_CTRLMMR_PADCONFIG + PIN_I2C1_SCL, PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_I2C1_SDA, PIN_INPUT_ENABLE | PIN_MODE(0));

    // pinmux for I2C3
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC2_CLK, PIN_INPUT_ENABLE | PIN_MODE(4));  // I2C3 SCL
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC2_CMD, PIN_INPUT_ENABLE | PIN_MODE(4));  // I2C3 SDA

    // pinmux for I2C6
    out32(J721E_CTRLMMR_PADCONFIG + PIN_SPI0_D1, PIN_INPUT_ENABLE | PIN_MODE(2));   // I2C6 SCL
    out32(J721E_CTRLMMR_PADCONFIG + PIN_SPI1_D1, PIN_INPUT_ENABLE | PIN_MODE(2));   // I2C6 SDA
}

static void init_gpu(void)
{
    if (ti_sci_set_device_state(TISCI_DEV_GPU0_GPU_0, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: Set GPU0 device state failed!\n", __func__);
    }
    if (ti_sci_set_device_state(TISCI_DEV_GPU0_GPUCORE_0, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: Set GPU1 device state failed!\n", __func__);
    }
}

static const uint8_t hdmi_pins[] = {
    PIN_PRG1_PRU0_GPO8,     // offset 0x24: VOUT0_DATA22
    PIN_PRG1_PRU0_GPO9,     // offset 0x28: VOUT0_DATA23
    PIN_PRG1_PRU0_GPO11,    // offset 0x30: VOUT0_DATA16
    PIN_PRG1_PRU0_GPO12,    // offset 0x34: VOUT0_DATA17
    PIN_PRG1_PRU0_GPO13,    // offset 0x38: VOUT0_DATA18
    PIN_PRG1_PRU0_GPO14,    // offset 0x3C: VOUT0_DATA19
    PIN_PRG1_PRU0_GPO15,    // offset 0x40: VOUT0_DATA20
    PIN_PRG1_PRU0_GPO16,    // offset 0x44: VOUT0_DATA21
    PIN_PRG1_PRU1_GPO0,     // offset 0x58: VOUT0_DATA0
    PIN_PRG1_PRU1_GPO1,     // offset 0x5C: VOUT0_DATA1
    PIN_PRG1_PRU1_GPO2,     // offset 0x60: VOUT0_DATA3
    PIN_PRG1_PRU1_GPO3,     // offset 0x64: VOUT0_DATA3
    PIN_PRG1_PRU1_GPO4,     // offset 0x68: VOUT0_DATA4
    PIN_PRG1_PRU1_GPO5,     // offset 0x6C: VOUT0_DATA5
    PIN_PRG1_PRU1_GPO6,     // offset 0x70: VOUT0_DATA6
    PIN_PRG1_PRU1_GPO7,     // offset 0x74: VOUT0_DATA7
    PIN_PRG1_PRU1_GPO8,     // offset 0x78: VOUT0_DATA8
    PIN_PRG1_PRU1_GPO9,     // offset 0x7C: VOUT0_DATA9
    PIN_PRG1_PRU1_GPO10,    // offset 0x80: VOUT0_DATA10
    PIN_PRG1_PRU1_GPO11,    // offset 0x84: VOUT0_DATA11
    PIN_PRG1_PRU1_GPO12,    // offset 0x88: VOUT0_DATA12
    PIN_PRG1_PRU1_GPO13,    // offset 0x8C: VOUT0_DATA13
    PIN_PRG1_PRU1_GPO14,    // offset 0x90: VOUT0_DATA14
    PIN_PRG1_PRU1_GPO15,    // offset 0x94: VOUT0_DATA15
    PIN_PRG1_PRU1_GPO16,    // offset 0x98: VOUT0_HSYNC
    PIN_PRG1_PRU1_GPO17,    // offset 0x9C: VOUT0_DE
    PIN_PRG1_PRU1_GPO18,    // offset 0xA0: VOUT0_VSYNC
    PIN_PRG1_PRU1_GPO19     // offset 0xA4: VOUT0_PCLK
};

static int init_dss(void)
{
    // DSS power, functional clock
    if (ti_sci_set_device_state(TISCI_DEV_DSS0, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: Set DSS device state failed!\n", __func__);
        return -1;
    }
    if (ti_sci_set_clock_state(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_FUNC_CLK,
            0, MSG_CLOCK_SW_STATE_REQ) == -1) {
        kprintf("%s: Set DSS clock state failed!\n", __func__);
        return -1;
    }

    return 0;
}

static int init_hdmi(void)
{
    uint32_t    hdmi_pincfg = PIN_INPUT_ENABLE | PIN_MODE(10);
    uint8_t     pin;

    // Pinmux
    for (pin = 0; pin < sizeof(hdmi_pins) / sizeof(uint8_t); pin++) {
        out32(J721E_CTRLMMR_PADCONFIG + hdmi_pins[pin], hdmi_pincfg);
    }

    if (ti_sci_cmd_set_clk_parent(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_1_IN_2X_CLK,
            TISCI_DEV_DSS0_DSS_INST0_DPI_1_IN_2X_CLK_PARENT_DPI0_EXT_CLKSEL_OUT0) == -1) {
        kprintf("%s: Set DSS clock parent failed!\n", __func__);
        return -1;
    }

    // Set 148.5Mhz pixel clock for 1080P output
    if (ti_sci_cmd_clk_set_freq(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_1_IN_2X_CLK,
            148500000ULL, 148500000ULL, 148500000ULL) == -1) {
        kprintf("%s: Set DSS pixel clock frequency failed!\n", __func__);
        return -1;
    }
    if (ti_sci_set_clock_state(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_1_IN_2X_CLK,
            0, MSG_CLOCK_SW_STATE_REQ) == -1) {
        kprintf("%s: Set DSS clock state failed!\n", __func__);
        return -1;
    }

    return 0;
}

static int init_dp(void)
{
    // Set HPD pin
    out32(J721E_CTRLMMR_PADCONFIG + PIN_SPI0_CS1, PIN_INPUT_ENABLE | PIN_MODE(5));

    // Enable DSS EDP0
    if (ti_sci_set_device_state(TISCI_DEV_DSS_EDP0, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: Set DSS DP device state failed!\n", __func__);
        return -1;
    }
    // Enable SERDES 10G0
    if (ti_sci_set_device_state(TISCI_DEV_SERDES_10G0, 0, MSG_DEVICE_SW_STATE_ON) == -1) {
        kprintf("%s: Set SERDES device state failed!\n", __func__);
        return -1;
    }

    if (ti_sci_cmd_set_clk_parent(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_2_IN_2X_CLK,
            TISCI_DEV_DSS0_DSS_INST0_DPI_2_IN_2X_CLK_PARENT_HSDIV1_16FFT_MAIN_18_HSDIVOUT0_CLK) == -1) {
        kprintf("%s: Set DSS clock parent failed!\n", __func__);
        return -1;
    }
    if (ti_sci_cmd_set_clk_parent(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_3_IN_2X_CLK,
            TISCI_DEV_DSS0_DSS_INST0_DPI_3_IN_2X_CLK_PARENT_DPI1_EXT_CLKSEL_OUT0) == -1) {
        kprintf("%s: Set DSS clock parent failed!\n", __func__);
        return -1;
    }
    if (ti_sci_cmd_set_clk_parent(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_0_IN_2X_CLK,
            TISCI_DEV_DSS0_DSS_INST0_DPI_0_IN_2X_CLK_PARENT_HSDIV1_16FFT_MAIN_16_HSDIVOUT0_CLK) == -1) {
        kprintf("%s: Set DSS clock parent failed!\n", __func__);
        return -1;
    }

    // Set 119Mhz pixel clock for 1680x1050 output
    if (ti_sci_cmd_clk_set_freq(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_0_IN_2X_CLK,
            119000000ULL, 119000000ULL, 119000000ULL) == -1) {
        kprintf("%s: Set DSS pixel clock frequency failed!\n", __func__);
        return -1;
    }
    if (ti_sci_set_clock_state(TISCI_DEV_DSS0, TISCI_DEV_DSS0_DSS_INST0_DPI_0_IN_2X_CLK,
            0, MSG_CLOCK_SW_STATE_REQ) == -1) {
        kprintf("%s: Set DSS clock state failed!\n", __func__);
        return -1;
    }

    return 0;
}

static const uint8_t ospi_pins[] = {
    PIN_MCU_OSPI0_CLK,
//    PIN_MCU_OSPI0_LBCLKO,
    PIN_MCU_OSPI0_DQS,
    PIN_MCU_OSPI0_D0,
    PIN_MCU_OSPI0_D1,
    PIN_MCU_OSPI0_D2,
    PIN_MCU_OSPI0_D3,
    PIN_MCU_OSPI0_D4,
    PIN_MCU_OSPI0_D5,
    PIN_MCU_OSPI0_D6,
    PIN_MCU_OSPI0_D7,
    PIN_MCU_OSPI0_CSN0
};

static void init_ospi(void)
{
    paddr_t		padcfg_base = J721E_WKUP_CTRLMMR_PADCONFIG;
	uint8_t		pin;

    // Set reference clock parent
    if (ti_sci_cmd_set_clk_parent(TISCI_DEV_MCU_FSS0_OSPI_0, TISCI_DEV_MCU_FSS0_OSPI_0_OSPI_RCLK_CLK,
            TISCI_DEV_MCU_FSS0_OSPI_0_OSPI_RCLK_CLK_PARENT_HSDIV4_16FFT_MCU_2_HSDIVOUT4_CLK) == -1) {
        kprintf("%s: Set OSPI0 RCLK clock parent failed!\n", __func__);
        return;
    }

    // Set 166.6Mhz reference clock
    if (ti_sci_cmd_clk_set_freq(TISCI_DEV_MCU_FSS0_OSPI_0, TISCI_DEV_MCU_FSS0_OSPI_0_OSPI_RCLK_CLK,
            166666666ULL, 166666666ULL, 166666666ULL) == -1) {
        kprintf("%s: Set OSPI0 RCLK frequency failed!\n", __func__);
        return;
    }

    // Pinmux
    for (pin = 0; pin < sizeof(ospi_pins) / sizeof(uint8_t); pin++) {
        if ((ospi_pins[pin] == PIN_MCU_OSPI0_CLK) || (ospi_pins[pin] == PIN_MCU_OSPI0_CSN0)) {
            out32(padcfg_base + ospi_pins[pin], PIN_MODE(0));
        } else {
            out32(padcfg_base + ospi_pins[pin], PIN_INPUT_ENABLE | PIN_MODE(0));
        }
    }
}

#define EVM_SDIO_1V8_EN_PIN_NUM     (117U)
static void init_sdmmc(void)
{
    // pinmux for SD(MMC1)
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_DAT3, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_DAT2, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_DAT1, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_DAT0, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_CLK, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    out32(J721E_CTRLMMR_PADCONFIG + PIN_MMC1_CMD, PIN_PULL_DISABLE | PIN_INPUT_ENABLE | PIN_MODE(0));
    // GPIO0_117 for signal voltage switch
    out32(J721E_CTRLMMR_PADCONFIG + PIN_SPI1_CS1, PIN_PULL_DISABLE | PIN_MODE(7));
    // GPIO0_117 as output
    out32(J721E_GPIO0_BASE + J721E_GPIO_DIR(EVM_SDIO_1V8_EN_PIN_NUM),
        in32(J721E_GPIO0_BASE + J721E_GPIO_DIR(EVM_SDIO_1V8_EN_PIN_NUM)) & ~(J721E_GPIO_BIT(EVM_SDIO_1V8_EN_PIN_NUM)));
    // Default signal voltage 3.3v, output high
    out32(J721E_GPIO0_BASE + J721E_GPIO_SET_DATA(EVM_SDIO_1V8_EN_PIN_NUM), J721E_GPIO_BIT(EVM_SDIO_1V8_EN_PIN_NUM));
}

/*
 * hw_init()
 *    Board specific initialization
 */
void
hw_init(void)
{
    init_i2c();
    init_gpu();
    if (r5_display == 0) {
        kprintf("Display set to A72 DSS\n");
        init_dss();
        init_hdmi();
        init_dp();
    }
    else {
        kprintf("Display set to R5\n");
    }
    init_ospi();
    init_sdmmc();

    init_usb_host(0, 1);
    init_usb_host(1, 1);
}

#if defined(__QNXNTO__) && defined(__USESRCVERSION)
#include <sys/srcversion.h>
__SRCVERSION("$URL: http://svn.ott.qnx.com/product/branches/7.1.0/trunk/hardware/startup/boards/j721e/evm/hw_init.c $ $Rev: 917234 $")
#endif
